from leaf_playground.core.scene_definition import *


SCENE_DEFINITION = SceneDefinition(
    name="{{ cookiecutter.project_name_camel_case }}",
    description="",
    roles=[],
    env_vars=[]
)


__all__ = [
    "SCENE_DEFINITION"
]
